#include "consola.h"
#include "klib.h"
#include "i386.h"
#include "loader.h"
#include "kbd.h"

consola_t consolas[MAX_CONSOLAS];
int consola_activa = -1;

static char reloj[] = "/-\\|";
static int reloj_pos = 0;

static void mover_cursor(int x, int y) {
	unsigned temp = CONS_ANCHO * y + x;

	/* This sends a command to indicies 14 and 15 in the
	*  CRT Control Register of the VGA controller. These
	*  are the high and low bytes of the index that show
	*  where the hardware cursor is to be 'blinking'. To
	*  learn more, you should look up some VGA specific
	*  programming documents. A great start to graphics:
	*  http://www.brackeen.com/home/vga */
	outb(0x3D4, 14);
	outb(0x3D5, temp >> 8);
	outb(0x3D4, 15);
	outb(0x3D5, temp);
}

// Codigo principal

void cons_inicializar() {
	for (int i = 0; i < MAX_CONSOLAS; i++) {
		// FIXME: asumimos que salen contiguas al comienzo (antes de correr tareas),
		// mejor mapearlas mas adelante
		consolas[i].buffer = obtener_pagina_kernel();
		obtener_pagina_kernel();
		obtener_pagina_kernel();
		obtener_pagina_kernel();
		memsetw(consolas[i].buffer, CHAR_BLANK, CONS_MAX_ELEMS);
		consolas[i].cursor = 0;
		consolas[i].scroll = 0;
		consolas[i].flags = 0;
		consolas[i].tarea = -1;
		consolas[i].input[0] = '\0';
		consolas[i].input_pos = 0;
	}
	cons_cambiar(0);
}

void cons_cambiar(int num_cons) {
	consola_t *cons = &consolas[num_cons];
	char buf[40];

	if (consola_activa == num_cons)
		return;

	consola_activa = num_cons;

	snprintf(buf, sizeof(buf), "  ZZ Top (consola %d)", num_cons);
	scr_writeln(buf, COL_BANDA1, 0, 0);
	scr_writeln("", COL_BANDA2, 1, 0);
	scr_writeln("", COL_BANDA2, 23, 0);
	scr_writeln("", COL_BANDA1, 24, 0);
	scr_write("Clock: *", COL_BANDA2, 23, 70);

	SET_ADD(cons->flags, CF_DIRTY);
}

void cons_putch(int num_cons, char ch, char modo) {
	kassert(num_cons >= 0 && num_cons < MAX_CONSOLAS);
	consola_t *cons = &consolas[num_cons];
	int tmp;

	switch (ch) {
	case '\n':
		tmp = CONS_ANCHO - CURSOR_X(cons);
		memsetw(&cons->buffer[cons->cursor], ' ' | (modo << 8), tmp);
		DESPLAZAR_CURSOR(cons, tmp);
		break;
	case '\r':
		DESPLAZAR_CURSOR(cons, -CURSOR_X(cons));
		break;
	case '\b':
		DESPLAZAR_CURSOR(cons, -1);
		break;
	case '\t':
		DESPLAZAR_CURSOR(cons, 8 - (CURSOR_X(cons) % 8));
		if (CURSOR_X(cons) == 0)
			memsetw(&cons->buffer[cons->cursor], ' ' | (modo << 8), CONS_ANCHO);
		break;
	default:
		if (ch >= ' ')
			cons->buffer[cons->cursor] = ch | (modo << 8);
		else
			cons->buffer[cons->cursor] = '?' | (COLOR(FG_LIGHT(COL_GRIS), COL_ROJO) << 8);
		DESPLAZAR_CURSOR(cons, 1);
		if (CURSOR_X(cons) == 0)
			memsetw(&cons->buffer[cons->cursor], CHAR_BLANK, CONS_ANCHO);
	}

	SET_ADD(cons->flags, CF_DIRTY);
}

int cons_putstr(int num_cons, const char *msg, char modo) {
	int msglen = strlen(msg);
	for (int i = 0; i < msglen; i++)
		cons_putch(num_cons, msg[i], modo);
	return msglen;
}

int cons_printf(int num_cons, char modo, const char *format, ...) {
	int ret;
	va_list ap;

	va_start(ap, format);
	ret = cons_vprintf(num_cons, modo, format, ap);
	va_end(ap);

	return ret;
}

int cons_vprintf(int num_cons, char modo, const char *format, va_list ap) {
	char buffer[128];
	int ret;

	ret = vsnprintf(buffer, 128, format, ap);
	cons_putstr(num_cons, buffer, modo);

	return ret;
}

// Para ser llamada por tareas, no kernel
int cons_ngets(char *buf, int size) {
	int num_cons = getcons(), pid = getpid();
	kassert(num_cons >= 0 && num_cons < MAX_CONSOLAS);
	consola_t *cons = &consolas[num_cons];
	int res;

	kassert(!SET_TEST(cons->flags, CF_ENTRADA));
	SET_ADD(cons->flags, CF_ENTRADA);
	sched_block(pid);

	strlcpy(buf, cons->input, size);
	res = cons->input_pos;

	// Resetear la entrada y mostrar lo que realmente se leyo
	cons->input[0] = '\0';
	cons->input_pos = 0;
	cons_printf(num_cons, cons->buffer[cons->cursor-1] >> 8, "%s\n", buf);
	cons_actualizar();	// XXX: borra el texto de la barra inferior, que no se toca luego
	SET_REMOVE(cons->flags, CF_ENTRADA);

	return res;
}

int cons_obtenerNueva(int pid) {
	for (int i = 0; i < MAX_CONSOLAS; i++)
		if (consolas[i].tarea < 0) {
			consolas[i].tarea = pid;
			return i;
		}
	kpanic("No hay mas consolas libres\n");
	return -1;
}

void cons_liberarUsada(int num_cons) {
	kassert(num_cons >= 0 && num_cons < MAX_CONSOLAS && consolas[num_cons].tarea >= 0);
	consolas[num_cons].tarea = -1;
	SET_REMOVE(consolas[num_cons].flags, CF_ENTRADA);
	SET_REMOVE(consolas[num_cons].flags, CF_SCROLLING);
}

void cons_limpiar(int num_cons) {
	kassert(num_cons >= 0 && num_cons < MAX_CONSOLAS);
	consola_t *cons = &consolas[num_cons];
	for (int i = 0; i < CONS_TEXTO_ALTO + 1; i++)
		cons_putch(num_cons, '\n', MODO_DEFAULT);
	cons->scroll -= CONS_TEXTO_AREA;
	cons->cursor -= CONS_TEXTO_AREA;
}


/*
 * Actualiza la parte central, ya sea por cons_putch() o scrolling.
 * Y tambien la linea de entrada.
 */
void cons_actualizar() {
	consola_t *cons = &consolas[consola_activa];
	short *screen = (short*)0xB8000;

	// Actualizamos el reloj de abajo a la derecha
	reloj_pos = (reloj_pos + 1) % sizeof(reloj)/sizeof(reloj[0]);
	scr_repch(reloj[reloj_pos], 23, 77);

	// No actualizamos el resto si no es necesario
	if (!SET_TEST(cons->flags, CF_DIRTY))
		return;

	SET_REMOVE(cons->flags, CF_DIRTY);

	if (!SET_TEST(cons->flags, CF_SCROLLING)) {
		// Actualizar cons->scroll, por si el cursor se adelanto
		if (CURSOR_Y(cons) >= CONS_TEXTO_ALTO)
			cons->scroll = (cons->scroll + (CURSOR_Y(cons) - CONS_TEXTO_ALTO + 1) * CONS_ANCHO) % CONS_MAX_ELEMS;
		mover_cursor(CURSOR_X(cons), CONS_MARGEN_TOP + CURSOR_Y(cons));
	}
	
	circular_read(cons->buffer, 2*CONS_MAX_ELEMS, 2*cons->scroll,
		    2*CONS_TEXTO_AREA, &screen[CONS_MARGEN_TOP*CONS_ANCHO]);

	// En modo "entrada" volvemos a escribir la linea, y llevamos el cursor alli
	if (SET_TEST(cons->flags, CF_ENTRADA)) {
		scr_writeln(cons->input, COL_BANDA1, 24, 0);
		mover_cursor(cons->input_pos, 24);
	}
}

void cons_haytecla(unsigned ch) {
	consola_t *cons = &consolas[consola_activa];

	// F1 a F12 cambia entre consolas virtuales
	if (ch >= KEY_F1 && ch <= KEY_F12)
		cons_cambiar(ch - KEY_F1);

	// Page Up y Page Down se usan para hacer scrolling
	/*
	 * TODO: hacer que no se pueda bajar mas del cursor, ni retroceder hasta que se vea el mismo.
	 * En ese caso tampoco haria falta inicializar toda la consola con espacios, solo es espacio inicialmente visible.
	 */
	else if (ch == KEY_PGUP) {
		SET_ADD(cons->flags, CF_SCROLLING);
		SET_ADD(cons->flags, CF_DIRTY);
		// XXX: el modulo no anda como se espera con los numeros negativos!
		cons->scroll -= CONS_ANCHO*(CONS_TEXTO_ALTO/2);
		if (cons->scroll < 0)
			cons->scroll += CONS_MAX_ELEMS;
	} else if (ch == KEY_PGDOWN) {
		SET_ADD(cons->flags, CF_SCROLLING);
		SET_ADD(cons->flags, CF_DIRTY);
		cons->scroll = (cons->scroll + CONS_ANCHO*(CONS_TEXTO_ALTO/2)) % CONS_MAX_ELEMS;
	}

	// Con ESC dejamos de hacer scroll y volvemos a la ultima posicion escrita
	else if (ch == 27) {
		SET_REMOVE(cons->flags, CF_SCROLLING);
		SET_ADD(cons->flags, CF_DIRTY);
		cons->scroll = (cons->scroll + (CURSOR_Y(cons) - CONS_TEXTO_ALTO + 1) * CONS_ANCHO) % CONS_MAX_ELEMS;
	}

	// En modo entrada, reaccionamos ante caracteres imprimibles, ENTER y BACKSPACE
	else if (SET_TEST(cons->flags, CF_ENTRADA)) {
		if (ch == '\n') {
			sched_unblock(cons->tarea);
			// XXX: el resto se hace detro de cons_ngets()
		} else if (ch == '\b') {
			if (cons->input_pos > 0) {
				cons->input[--cons->input_pos] = '\0';
				SET_ADD(cons->flags, CF_DIRTY);
			}
		} else if (isprint(ch)) {
			if (cons->input_pos < CONS_MAX_LINE) {
				cons->input[cons->input_pos++] = ch;
				cons->input[cons->input_pos] = '\0';
				SET_ADD(cons->flags, CF_DIRTY);
			}
		}
	}
}
