; vim:ft=nasm
BITS 32

global _isr0,_isr1,_isr2,_isr3,_isr4,_isr5,_isr6,_isr7,_isr8,_isr9,_isr10
global _isr11,_isr12,_isr13,_isr14,_isr15,_isr16,_isr17,_isr18,_isr19
global _isrclock,_isrteclado,_isrsyscall,
extern kbd_readkey, writescr, proxima_tarea, syscalls_tabla, syscalls_cant, upanic

_isr0:
	cli
	push byte 0
	push byte 0
	jmp isr_common_stub

_isr1:
	cli
	push byte 0
	push byte 1
	jmp isr_common_stub

_isr2:
	cli
	push byte 0
	push byte 2
	jmp isr_common_stub

_isr3:
	cli
	push byte 0
	push byte 3
	jmp isr_common_stub

_isr4:
	cli
	push byte 0
	push byte 4
	jmp isr_common_stub

_isr5:
	cli
	push byte 0
	push byte 5
	jmp isr_common_stub

_isr6:
	cli
	push byte 0
	push byte 6
	jmp isr_common_stub

_isr7:
	cli
	push byte 0
	push byte 7
	jmp isr_common_stub

_isr8:
	cli
	; Ya hay codigo de error
	push byte 8
	jmp isr_common_stub

_isr9:
	cli
	push byte 0
	push byte 9
	jmp isr_common_stub

_isr10:
	cli
	; Ya hay codigo de error
	push byte 10
	jmp isr_common_stub

_isr11:
	cli
	; Ya hay codigo de error
	push byte 11
	jmp isr_common_stub

_isr12:
	cli
	; Ya hay codigo de error
	push byte 12
	jmp isr_common_stub

_isr13:
	cli
	; Ya hay codigo de error
	push byte 13
	jmp isr_common_stub

_isr14:
	cli
	; Ya hay codigo de error
	push byte 14
	jmp isr_common_stub

_isr15:
	cli
	push byte 0
	push byte 15
	jmp isr_common_stub

_isr16:
	cli
	push byte 0
	push byte 16
	jmp isr_common_stub

_isr17:
	cli
	push byte 0
	push byte 17
	jmp isr_common_stub

_isr18:
	cli
	push byte 0
	push byte 18
	jmp isr_common_stub

_isr19:
	cli
	push byte 0
	push byte 19
	jmp isr_common_stub

; This is our common ISR stub. It saves the processor state, sets
; up for kernel mode segments, calls the C-level fault handler,
; and finally restores the stack frame.
extern interrupt_handler
isr_common_stub:
	pushad
	mov	eax, esp	; Push us the stack
	push	eax
	mov	eax, interrupt_handler
	call	eax	; A special call, preserves the 'eip' register
	pop	eax
	popad
	add	esp, 8	; Cleans up the pushed error code and pushed ISR number
	iret		; pops 5 things at once: CS, EIP, EFLAGS, SS, and ESP!

_isrclock:
	cli
	pushad

	; Informo al pic1 que atendi la interrupcion
	mov 	al, 0x20
	out 	0x20, al

	; Obtener la proxima tarea a correr
	xor eax, eax
	str	ax
	push	eax
	call proxima_tarea
	add	esp, 4

	; Si es la misma, retornamos en vez de camibar
	cmp eax, 0
	je .sigue

	; Llamarla con un "far jump"
	push	eax
	push    0         ; offset is ignored
	jmp     dword far [esp]
	add     esp, 8    ; if you switch back to this task, remove the dwords pushed onto the stack

.sigue:
	popad
	sti
	iret

_isrteclado:
	cli
	pushad

	; Informo al pic1 que atendi la interrupcion
	mov 	al, 0x20
	out	0x20, al

	; Esta funcion se encarga de interpretarla y pasarla al modulo de consola
	call	kbd_readkey

	popad
	sti
	iret

_isrsyscall:
	cli
	push	ebp
	mov	ebp, esp

	; Guardar estos segun la convencion C
	push	ebx
	push	esi
	push	edi

	; Verificar que sea un indice valido
	cmp	eax, syscalls_cant
	jae	.error

	mov	ebx, syscalls_tabla
	mov	esi, [ebx+eax*8+4]	; cantidad de argumentos
	mov	ebx, [ebx+eax*8+0]	; direccion de la funcion
	mov	edi, [ebp+16]		; ESP de codigo de usuario (antes del syscall)
	add	edi, 8			; inicio del primer parametro (ultimo en agregarse)

	mov	ecx, esi		; preparar ECX para iterar sobre los argumentos
	jecxz	.continuar		; si no hay parametros, NO ENTRAR EN EL LOOP
	shl	esi, 2			; offset en la pila hasta el ultimo (primero en agregarse)
	add	edi, esi

	; Hacer la copia al reves, manteniendo la convencion C
.cargarParametros:
	sub	edi, 4
	push	dword [edi]
	loop	.cargarParametros
	
.continuar:
	; Llamar a la syscall, restaurar la pila y registros pero dejar EAX intacto
	; (con el posible valor de retorno de la anterior)
	call	ebx
	add	esp, esi

	; Restaurar estos segun la convencion C
	pop	edi
	pop	esi
	pop	ebx

	pop	ebp
	sti
	iret
.error:
	push eax
	push msgInvSysCall
	call upanic	; no retorna, ya que el proceso se termina

msgInvSysCall: db "Invalid syscall %d",0
