#include "i386.h"
#include "klib.h"
#include "consola.h"
#include "loader.h"
#include "pak.h"

void kpanic(const char *format, ...) {
	int num_cons = getcons();
	va_list ap;
	const int color = COLOR(FG_LIGHT(COL_ROJO), COL_NEGRO);

	cons_printf(num_cons, color,
	    "\nError de KERNEL ejecutando proceso %d en la consola %d:\n", getpid(), num_cons);
	if (format != NULL) {
		va_start(ap, format);
		cons_vprintf(num_cons, color, format, ap);
		va_end(ap);
	} else
		cons_printf(num_cons, color, "<sin descripcion>\n");
	cons_cambiar(num_cons);
	cons_actualizar();

	// TODO: terminar todas las tareas pero permitir hacer scrolling?
	// TODO: permitir hacer un backtrace? (si llamamos a kpanic sin excepcion no estamos en el horno)
	for (;;)
		;
}

void upanic(const char *format, ...) {
	int num_cons = getcons();
	int pid = getpid();
	va_list ap;
	const int color = COLOR(COL_VERDE, COL_NEGRO);

	cons_printf(num_cons, color,
	    "\nError de usuario ejecutando '%s' (PID %d) en la consola %d:\n",
	    procesos[pid].name, pid, num_cons);
	if (format != NULL) {
		va_start(ap, format);
		cons_vprintf(num_cons, color, format, ap);
		va_end(ap);
	} else
		cons_printf(num_cons, color, "<sin descripcion>\n");
	cons_cambiar(num_cons);
	sched_kill(getpid(), 0xDEAD);
}

void scr_write(const char* msg, unsigned format, unsigned row, unsigned col) {
	unsigned short* scr = SCREEN;
	scr += row * 80 + col;
	
	while (*msg)
		*scr++ = (format << 8) | *msg++;
}

void scr_writeln(const char* msg, unsigned format, unsigned row, unsigned col) {
	unsigned short* scr = SCREEN;
	int left = 80 - col;
	scr += row * 80 + col;
	
	while (*msg) {
		*scr++ = (format << 8) | *msg++;
		left--;
	}
	while (left-- > 0)
		*scr++ = (format << 8) | ' ';
}

void scr_putch(char ch, int format, int row, int col) {
	unsigned short* scr = SCREEN;
	scr[row * 80 + col] = (format << 8) | ch;
}

void scr_repch(char ch, int row, int col) {
	unsigned short* scr = SCREEN;
	char *pos = (char*) &scr[row * 80 + col];
	*pos = ch;
}

void circular_read(const void *buffer, int size, int offset, int ammount, void *to) {
	const char *src = (const char*)buffer;
	char *dst = (char*)to;

	while (ammount > 0) {
		if (offset >= size) {
			offset -= size;
			continue;
		}
		*dst++ = src[offset++];
		ammount--;
	}
}

int copyInPages(void **dest, int maxPages, const void *source, int size) {
	const char *srcPtr = source;
	int i;
	for (i = 0; i < maxPages && size > 0; i++) {
		int amt = MIN(size, TAMANO_PAGINA);
		memcpy(dest[i], srcPtr, amt);
		size -= amt;
		srcPtr += amt;
	}
	return (size > 0) ? -1 : i;
}

// System calls

void sys_puts(const char *s) {
	cons_putstr(getcons(), s, MODO_DEFAULT);
}

int sys_ngets(char *buf, int size) {
	return cons_ngets(buf, size);
}

void sys_exit(int status) {
	sched_kill(getpid(), status);
}

int sys_getpid() {
	return TR_A_PID(rtr());
}

int sys_getcons() {
	return procesos[getpid()].consola;
}

int sys_ls() {
	// Solo el shell principal puede imprimir la lista de tareas
	if (getpid() != PID_SHELL)
		return ERR_PERM;

	return pak_list(getcons());
}

int sys_exec(const char *name, int quantum) {
	int pid = getpid();

	// Solo el shell principal puede lanzar procesos
	if (pid != PID_SHELL)
		return ERR_PERM;
	if (num_procesos >= MAX_PROCESOS)
		return ERR_MAX;
	if (pak_find(name) == NULL)
		return ERR_NOENT;

	return loader_exec(name, quantum);
}

int sys_ps() {
	// Solo el shell principal puede imprimir los procesos
	if (getpid() != PID_SHELL)
		return ERR_PERM;

	cons_printf(getcons(), COLOR(COL_MARRON, COL_NEGRO),
	    "PID, Nombre, Consola, Quantum, Estado\n");
	int i;
	for (i = 0; i < MAX_PROCESOS; i++) {
		if (!SET_TEST(procesos[i].flags, SCH_PRESENTE) ||
		    SET_TEST(procesos[i].flags, SCH_OCULTA))
			continue;
		cons_printf(getcons(), COLOR(COL_MARRON, COL_NEGRO),
		    "%d, %s, %d, %d, %s\n", i, procesos[i].name, procesos[i].consola, procesos[i].quantum,
		    SET_TEST(procesos[i].flags, SCH_BLOQUEADA) ? "SLEEPING" : "RUNNING");
	}
	return i;
}

int sys_kill(int pid) {
	// Solo el shell principal puede cerrar otro proceso
	if (getpid() != PID_SHELL)
		return ERR_PERM;
	// XXX: seria una catastrofe que esto ocurriera!
	if (pid == PID_IDLE)
		return ERR_PERM;
	if (!(pid >= 0 && pid < MAX_PROCESOS && SET_TEST(procesos[pid].flags, SCH_PRESENTE)))
		return ERR_NOENT;

	sched_kill(pid, 0xBEE);
	return 0;
}
