#include "pak.h"
#include "consola.h"
#include "klib.h"

pak_header_t *progsPAK;

void pak_init() {
	pak_header_t *pak = (void*)0x10000;
	if (pak->magic != PAK_MAGIC)
		kpanic("Wrong PAK magic number: %x\n", pak->magic);
	progsPAK = pak;
}

const pak_entry_t* pak_find(const char *name) {
	const pak_header_t *pak = progsPAK;
	const pak_entry_t *pent = (const pak_entry_t*)(pak+1);
	for (int i = 0; i < pak->num_entries; i++)
		if (strncmp(pent[i].name, name, PAK_MAX_NAME) == 0)
			return &pent[i];
	return NULL;
}

void pak_copyTo(const char *name, void *dest) {
	const pak_header_t *pak = progsPAK;
	const pak_entry_t *pent = pak_find(name);
	kassert(pent != NULL);
	memcpy(dest, ((const char*)pak)+pent->position, pent->size);
}

int pak_copyToPages(const char *name, void **dest, int maxPages) {
	const pak_header_t *pak = progsPAK;
	const pak_entry_t *pent = pak_find(name);
	kassert(pent != NULL);
	int res = copyInPages(dest, maxPages, ((const char*)pak)+pent->position, pent->size);
	kassert(res > 0);
	return res;
}

int pak_list(int num_cons) {
	const pak_header_t *pak = progsPAK;
	const pak_entry_t *pent = (const pak_entry_t*)(pak+1);
	int i;

	cons_printf(num_cons, COLOR(COL_MARRON, COL_NEGRO),
	    "Nombre, Tamanio (en bytes)\n");
	for (i = 0; i < pak->num_entries; i++) {
		cons_printf(num_cons, COLOR(COL_MARRON, COL_NEGRO),
		    "%s, %d\n", pent[i].name, pent[i].size);
	}
	return i;
}
